# -*- coding: utf-8 -*-
## Description: class NetworkML for loading NetworkML from file or xml element into MOOSE
## Version 1.0 by Aditya Gilra, NCBS, Bangalore, India, 2011 for serial MOOSE
## Version 1.5 by Niraj Dudani, NCBS, Bangalore, India, 2012, ported to parallel MOOSE
## Version 1.6 by Aditya Gilra, NCBS, Bangalore, India, 2012, further changes for parallel MOOSE
## Version 1.7 by Aditya Gilra, NCBS, Bangalore, India, 2013, further support for NeuroML 1.8.1
## Version 1.8 by Aditya Gilra, NCBS, Bangalore, India, 2013, changes for new IntFire and SynHandler classes

"""
NeuroML.py is the preferred interface. Use this only if NeuroML L1,L2,L3 files are misnamed/scattered.
Instantiate NetworkML class, and thence use method:
readNetworkMLFromFile(...) to load a standalone NetworkML file, OR
readNetworkML(...) to load from an xml.etree xml element (could be part of a larger NeuroML file).
"""

from __future__ import print_function
from xml.etree import cElementTree as ET
import string
import os
from math import cos, sin
from moose.neuroml.MorphML import MorphML
from moose.neuroml.ChannelML import ChannelML, make_new_synapse
import moose
from moose.neuroml.utils import meta_ns, nml_ns, find_first_file, tweak_model
from moose import utils

import logging
# define a Handler which writes INFO messages or higher to the sys.stderr
console = logging.StreamHandler()
console.setLevel(logging.INFO)
# set a format which is simpler for console use
formatter = logging.Formatter('%(name)-12s: %(levelname)-8s %(message)s')
# tell the handler to use this format
console.setFormatter(formatter)
# add the handler to the root logger
logging.getLogger('moose.nml.networkml').addHandler(console)
_logger = logging.getLogger('moose.nml.networkml')

class NetworkML():

    def __init__(self, nml_params):
        self.nml_params = nml_params
        self.model_dir = nml_params['model_dir']

    def readNetworkMLFromFile(self,filename,cellSegmentDict,params={}):
        """
        specify tweak params = {'excludePopulations':[popname1,...], 'excludeProjections':[projname1,...], \
            'onlyInclude':{'includePopulation':(popname,[id1,...]),'includeProjections':(projname1,...)} }
        If excludePopulations is present, then excludeProjections must also be present:
        Thus if you exclude some populations,
            ensure that you exclude projections that refer to those populations also!
        Though for onlyInclude, you may specify only included cells and this reader will
            also keep cells connected to those in onlyInclude.
        This reader first prunes the exclude-s,
            then keeps the onlyInclude-s and those that are connected.
        Use 'includeProjections' if you want to keep some projections not connected to
            the primary 'includePopulation' cells
        but connected to secondary cells that connected to the primary ones:
        e.g. baseline synapses on granule cells connected to 'includePopulation' mitrals;
            these synapses receive file based pre-synaptic events,
            not presynaptically connected to a cell.
        In params, you further specify:
            'createPotentialSynapses' : True (False by default)
            to create synapses at all potential locations/compartments specified in the MorphML cell file
            even before Projections tag is parsed.
            'combineSegments' : True (False by default)
            to ask neuroml to combine segments belonging to a cable
            (Neuron generates multiple segments per section).
        """
        _logger.info("Reading file %s " % filename)
        tree = ET.parse(filename)
        root_element = tree.getroot()
        _logger.info("Tweaking model ... ")
        tweak_model(root_element, params)
        _logger.info("Loading model into MOOSE ... ")
        return self.readNetworkML(root_element,cellSegmentDict,params,root_element.attrib['lengthUnits'])

    def readNetworkML(self,network,cellSegmentDict,params={},lengthUnits="micrometer"):
        """
        This returns populationDict = { 'populationname1':(cellname,{int(instanceid1):moosecell, ... }) , ... }
        and projectionDict = { 'projectionname1':(source,target,[(syn_name1,pre_seg_path,post_seg_path),...]) , ... }
        """
        if lengthUnits in ['micrometer','micron']:
            self.length_factor = 1e-6
        else:
            self.length_factor = 1.0
        self.network = network
        self.cellSegmentDict = cellSegmentDict
        self.params = params

        self.populationDict = {}
        [ self.createPopulation(pop) for pop in
                self.network.findall(".//{"+nml_ns+"}population")
                ]

        self.projectionDict={}
        projections = self.network.find(".//{"+nml_ns+"}projections")
        if projections:
            # see pg 219 (sec 13.2) of Book of Genesis
            if projections.attrib["units"] == 'Physiological Units':
                Efactor = 1e-3 # V from mV
                Tfactor = 1e-3 # s from ms
            else:
                Efactor = 1.0
                Tfactor = 1.0
            [ self.createProjection(proj, Efactor, Tfactor) for proj in projections ]

        allinputs = self.network.findall(".//{"+nml_ns+"}inputs")
        for inputs in allinputs:
            _logger.info("Creating input under /elec ")
            units = inputs.attrib['units']
            # see pg 219 (sec 13.2) of Book of Genesis
            if units == 'Physiological Units':
                Vfactor, Tfactor, Ifactor = 1e-3, 1e-3, 1e-6
            else:
                Vfactor, Tfactor, Ifactor = 1.0, 1.0, 1.0
            [ self.createInput(inputelem, Vfactor, Tfactor, Ifactor) for
                    inputelem in self.network.findall(".//{"+nml_ns+"}input")
                    ]

        return (self.populationDict,self.projectionDict)

    def createInput(self, inputelem, Vfactor, Tfactor, Ifactor):
        """Create input """
        inputname = inputelem.attrib['name']
        pulseinput = inputelem.find(".//{"+nml_ns+"}pulse_input")
        if pulseinput is not None:
            ## If /elec doesn't exists it creates /elec
            ## and returns a reference to it. If it does,
            ## it just returns its reference.
            moose.Neutral('/elec')
            pulsegen = moose.PulseGen('/elec/pulsegen_'+inputname)
            iclamp = moose.DiffAmp('/elec/iclamp_'+inputname)
            iclamp.saturation = 1e6
            iclamp.gain = 1.0
            pulsegen.trigMode = 0 # free run
            pulsegen.baseLevel = 0.0
            _logger.debug("Tfactor, Ifactor: %s, %s" % (Tfactor, Ifactor))
            _logger.debug("Pulsegen attributes: %s" % str(pulseinput.attrib))
            pulsegen.firstDelay = float(pulseinput.attrib['delay'])*Tfactor
            pulsegen.firstWidth = float(pulseinput.attrib['duration'])*Tfactor
            pulsegen.firstLevel = float(pulseinput.attrib['amplitude'])*Ifactor
            pulsegen.secondDelay = 1e6 # to avoid repeat
            pulsegen.secondLevel = 0.0
            pulsegen.secondWidth = 0.0
            ## do not set count to 1, let it be at 2 by default
            ## else it will set secondDelay to 0.0 and repeat the first pulse!
            #pulsegen.count = 1
            moose.connect(pulsegen,'output',iclamp,'plusIn')
            target = inputelem.find(".//{"+nml_ns+"}target")
            population = target.attrib['population']
            for site in target.findall(".//{"+nml_ns+"}site"):
                cell_id = site.attrib['cell_id']
                if 'segment_id' in site.attrib: segment_id = site.attrib['segment_id']
                else: segment_id = 0 # default segment_id is specified to be 0
                ## population is populationname, self.populationDict[population][0] is cellname
                cell_name = self.populationDict[population][0]
                segment_path = self.populationDict[population][1][int(cell_id)].path+'/'+\
                    self.cellSegmentDict[cell_name][0][segment_id][0]
                compartment = moose.element(segment_path)
                _logger.debug("Adding pulse at {0}: {1}".format(
                    segment_path, pulsegen.firstLevel )
                    )

                _logger.debug("Connecting {0}:output to {1}:injectMst".format(
                    iclamp, compartment)
                    )

                moose.connect(iclamp, 'output', compartment, 'injectMsg')

    def createPopulation(self, population):
        """Create a population with given cell type """
        cellname = population.attrib["cell_type"]
        populationname = population.attrib["name"]

        if not moose.exists('/library/'+cellname):
            ## if cell does not exist in library load it from xml file
            mmlR = MorphML(self.nml_params)
            model_filenames = (cellname+'.xml', cellname+'.morph.xml')
            success = False
            for model_filename in model_filenames:
                model_path = find_first_file(model_filename, self.model_dir)
                if model_path is not None:
                    cellDict = mmlR.readMorphMLFromFile(model_path, self.params)
                    success = True
                    break
            if not success:
                raise IOError(
                    'For cell {0}: files {1} not found under {2}.'.format(
                        cellname, model_filenames, self.model_dir
                    )
                )
            self.cellSegmentDict.update(cellDict)

        libcell = moose.Neuron('/library/'+cellname) #added cells as a Neuron class.
        self.populationDict[populationname] = (cellname,{})
        moose.Neutral('/cells')
        _logger.info(
                "Creating population {0} of cell type {1}".format(
                    populationname, cellname
                    )
                )

        for instance in population.findall(".//{"+nml_ns+"}instance"):
            instanceid = instance.attrib['id']
            location = instance.find('./{'+nml_ns+'}location')
            rotationnote = instance.find('./{'+meta_ns+'}notes')
            if rotationnote is not None:
                ## the text in rotationnote is zrotation=xxxxxxx
                zrotation = float(rotationnote.text.split('=')[1])
            else:
                zrotation = 0
            ## deep copies the library cell to an instance under '/cells' named as <arg3>
            ## /cells is useful for scheduling clocks as all sim elements are in /cells
            cellid = moose.copy(libcell,moose.Neutral('/cells'),populationname+"_"+instanceid)
            cell = moose.Neuron(cellid)
            self.populationDict[populationname][1][int(instanceid)]=cell
            x = float(location.attrib['x'])*self.length_factor
            y = float(location.attrib['y'])*self.length_factor
            z = float(location.attrib['z'])*self.length_factor
            self.translate_rotate(cell,x,y,z,zrotation)

    def translate_rotate(self,obj,x,y,z,ztheta): # recursively translate all compartments under obj
        for childId in obj.children:
            try:
                childobj = moose.element(childId)
                if childobj.className in ['Compartment','SymCompartment']:
                    ## SymCompartment inherits from Compartment,
                    ## so below wrapping by Compartment() is fine for both Compartment and SymCompartment
                    child = moose.element(childId)
                    x0 = child.x0
                    y0 = child.y0
                    x0new = x0*cos(ztheta)-y0*sin(ztheta)
                    y0new = x0*sin(ztheta)+y0*cos(ztheta)
                    child.x0 = x0new + x
                    child.y0 = y0new + y
                    child.z0 += z
                    x1 = child.x
                    y1 = child.y
                    x1new = x1*cos(ztheta)-y1*sin(ztheta)
                    y1new = x1*sin(ztheta)+y1*cos(ztheta)
                    child.x = x1new + x
                    child.y = y1new + y
                    child.z += z
                if len(childobj.children)>0:
                    self.translate_rotate(childobj,x,y,z,ztheta) # recursive translation+rotation
            except TypeError:  # in async13, gates which have not been created still 'exist'
                                # i.e. show up as a child, but cannot be wrapped.
                pass
            ## if childobj is a compartment or symcompartment translate, else skip it

    def createProjection(self, projection, Efactor, Tfactor):
        projectionname = projection.attrib["name"]
        _logger.info("Setting %s" % projectionname)
        source = projection.attrib["source"]
        target = projection.attrib["target"]
        self.projectionDict[projectionname] = (source,target,[])
        for syn_props in projection.findall(".//{"+nml_ns+"}synapse_props"):
            syn_name = syn_props.attrib['synapse_type']
            ## if synapse does not exist in library load it from xml file
            if not moose.exists("/library/"+syn_name):
                cmlR = ChannelML(self.nml_params)
                model_filename = syn_name+'.xml'
                model_path = find_first_file(model_filename,self.model_dir)
                if model_path is not None:
                    cmlR.readChannelMLFromFile(model_path)
                else:
                    raise IOError(
                        'For mechanism {0}: files {1} not found under {2}.'.format(
                            syn_name, model_filename, self.model_dir
                        )
                    )
            weight = float(syn_props.attrib['weight'])
            threshold = float(syn_props.attrib['threshold'])*Efactor
            if 'prop_delay' in syn_props.attrib:
                prop_delay = float(syn_props.attrib['prop_delay'])*Tfactor
            elif 'internal_delay' in syn_props.attrib:
                prop_delay = float(syn_props.attrib['internal_delay'])*Tfactor
            else: prop_delay = 0.0
            for connection in projection.findall(".//{"+nml_ns+"}connection"):
                pre_cell_id = connection.attrib['pre_cell_id']
                post_cell_id = connection.attrib['post_cell_id']
                if 'file' not in pre_cell_id:
                    # source could be 'mitrals', self.populationDict[source][0] would be 'mitral'
                    pre_cell_name = self.populationDict[source][0]
                    if 'pre_segment_id' in connection.attrib:
                        pre_segment_id = connection.attrib['pre_segment_id']
                    else: pre_segment_id = "0" # assume default segment 0, usually soma
                    pre_segment_path = self.populationDict[source][1][int(pre_cell_id)].path+'/'+\
                        self.cellSegmentDict[pre_cell_name][0][pre_segment_id][0]
                else:
                    # I've removed extra excitation provided via files, so below comment doesn't apply.
                    # 'file[+<glomnum>]_<filenumber>' # glomnum is
                    # for mitral_granule extra excitation from unmodelled sisters.
                    pre_segment_path = pre_cell_id+'_'+connection.attrib['pre_segment_id']
                # target could be 'PGs', self.populationDict[target][0] would be 'PG'
                post_cell_name = self.populationDict[target][0]
                if 'post_segment_id' in connection.attrib:
                    post_segment_id = connection.attrib['post_segment_id']
                else: post_segment_id = "0" # assume default segment 0, usually soma
                post_segment_path = self.populationDict[target][1][int(post_cell_id)].path+'/'+\
                    self.cellSegmentDict[post_cell_name][0][post_segment_id][0]
                self.projectionDict[projectionname][2].append((syn_name, pre_segment_path, post_segment_path))
                properties = connection.findall('./{'+nml_ns+'}properties')
                if len(properties)==0:
                    self.connect(syn_name, pre_segment_path, post_segment_path, weight, threshold, prop_delay)
                else:
                    for props in properties:
                        synapse_type = props.attrib['synapse_type']
                        if syn_name in synapse_type:
                            weight_override = float(props.attrib['weight'])
                            if 'internal_delay' in props.attrib:
                                delay_override = float(props.attrib['internal_delay'])
                            else: delay_override = prop_delay
                            if weight_override != 0.0:
                                self.connect(syn_name, pre_segment_path, post_segment_path,\
                                    weight_override, threshold, delay_override)

    def connect(self, syn_name, pre_path, post_path, weight, threshold, delay):
        postcomp = moose.element(post_path)
        ## We usually try to reuse an existing SynChan & SynHandler -
        ## event based SynHandlers have an array of weights and delays and can represent multiple synapses,
        ## so a new element of the weights and delays array is created
        ## every time a 'synapse' message connects to the SynHandler (from 'event' of spikegen)
        ## BUT for a graded synapse with a lookup table output connected to 'activation' message,
        ## not to 'synapse' message, we make a new synapse everytime
        ## ALSO for a saturating synapse i.e. KinSynChan, we always make a new synapse
        ## as KinSynChan is not meant to represent multiple synapses
        libsyn = moose.SynChan('/library/'+syn_name)
        gradedchild = utils.get_child_Mstring(libsyn,'graded')
        if libsyn.className == 'KinSynChan' or gradedchild.value == 'True': # create a new synapse
            syn_name_full = syn_name+'_'+utils.underscorize(pre_path)
            make_new_synapse(syn_name, postcomp, syn_name_full, self.nml_params)
        else:
            ## if syn doesn't exist in this compartment, create it
            syn_name_full = syn_name
            if not moose.exists(post_path+'/'+syn_name_full):
                make_new_synapse(syn_name, postcomp, syn_name_full, self.nml_params)
        ## moose.element is a function that checks if path exists,
        ## and returns the correct object, here SynChan
        syn = moose.element(post_path+'/'+syn_name_full) # wrap the SynChan in this compartment
        synhandler = moose.element(post_path+'/'+syn_name_full+'/handler') # wrap the SynHandler
        gradedchild = utils.get_child_Mstring(syn,'graded')
        #### weights are set at the end according to whether the synapse is graded or event-based

        #### If graded, connect pre-comp Vm to the table which is connected to SynChan's activation
        #### If event-based, connect spikegen/timetable's spikeOut to Simple/STDP SynHandler's addSpike
        ## I rely on second term below not being evaluated if first term is None;
        ## otherwise None.value gives error.
        if gradedchild is not None and gradedchild.value=='True': # graded synapse
            interpol = moose.element(syn.path+"/graded_table")
            #### always connect source to input - else 'cannot create message' error.
            precomp = moose.element(pre_path)
            moose.connect(precomp,"VmOut",interpol,"input")
            try:
                tau_table = moose.element(syn.path+'/tau_table')
                tau_table_present = True
            except ValueError:
                tau_table_present = False
            # if tau_table is not found, don't connect it
            if tau_table_present:
                moose.connect(precomp,'VmOut',tau_table,'input')
            ## since there is no weight field for a graded synapse
            ## (no 'synapse' message connected),
            ## I set the Gbar to weight*Gbar
            syn.Gbar = weight*syn.Gbar
        else: # Event based synapse
            ## synapse could be connected to either spikegen at pre-compartment OR to a file!
            if 'file' not in pre_path:
                ## element() can return either Compartment() or IzhikevichNrn(),
                ## since it queries and wraps the actual object
                precomp = moose.element(pre_path)
                ## if spikegen for this synapse doesn't exist in this compartment, create it
                ## spikegens for different synapse_types can have different thresholds
                ## but an integrate and fire spikegen supercedes all other spikegens
                if 'IF' in precomp.className: # intfire LIF
                    spikegen = precomp # LIF has a spikeOut message
                else:
                    if not moose.exists(pre_path+'/'+syn_name+'_spikegen'):
                        ## create new spikegen
                        spikegen = moose.element(pre_path+'/'+syn_name+'_spikegen')
                        ## connect the compartment Vm to the spikegen
                        moose.connect(precomp,"VmOut",spikegen,"Vm")
                        ## spikegens for different synapse_types can have different thresholds
                        spikegen.threshold = threshold
                        spikegen.edgeTriggered = 1 # This ensures that spike is generated only on leading edge.
                        ## usually events are raised at every time step that Vm > Threshold,
                        ## can set either edgeTriggered as above or refractT
                        #spikegen.refractT = 0.25e-3
                    ## wrap the existing or newly created spikegen in this compartment
                    spikegen = moose.SpikeGen(pre_path+'/'+syn_name+'_spikegen')
                ## connect the spikegen to the SynHandler
                ## note that you need to use Synapse (auto-created) under SynHandler
                ## to get/set weights , addSpike-s etc.
                ## wrap Synapse element by moose.Synapse(synhandler.path+'/synapse') or synhandler.synapse
                ## Synpase is an array element, first add to it, to addSpike-s, get/set weights, etc.
                synhandler.numSynapses += 1
                ## see Demos/snippets/synapse.py for an example of
                ## how to connect multiple SpikeGens to the same SynChan
                m = moose.connect(spikegen, 'spikeOut',
                                    synhandler.synapse[-1], 'addSpike', 'Single')
            else:
                ## if connected to a file, create a timetable,
                ## put in a field specifying the connected filenumbers to this segment,
                ## and leave it for simulation-time connection
                ## pre_path is 'file[+<glomnum>]_<filenum1>[_<filenum2>...]' i.e. glomnum could be present
                ## hack for my (Aditya's) OB model to use files in NeuroML, should not affect others
                filesplit = pre_path.split('+')
                if len(filesplit) == 2:
                    glomsplit = filesplit[1].split('_',1)
                    glomstr = '_'+glomsplit[0]
                    filenums = glomsplit[1]
                else:
                    glomstr = ''
                    filenums = pre_path.split('_',1)[1]
                tt_path = postcomp.path+'/'+syn_name_full+glomstr+'_tt'
                if not moose.exists(tt_path):
                    ## if timetable for this synapse doesn't exist in this compartment, create it,
                    ## and add the field 'fileNumbers'
                    tt = moose.TimeTable(tt_path)
                    tt_filenums = moose.Mstring(tt_path+'/fileNumbers')
                    tt_filenums.value = filenums
                    ## Be careful to connect the timetable only once while creating it as below:
                    ## note that you need to use Synapse (auto-created) under SynChan
                    ## to get/set weights , addSpike-s etc.
                    ## wrap Synapse element by moose.Synapse(synhandler.path+'/synapse') or synhandler.synapse
                    ## Synpase is an array element, first add to it, to addSpike-s, get/set weights, etc.
                    synhandler.numSynapses += 1
                    m = moose.connect(tt,"eventOut",synhandler.synapse[-1],"addSpike","Single")
                else:
                    ## if it exists, append file number to the field 'fileNumbers'
                    ## append filenumbers from 'file[+<glomnum>]_<filenumber1>[_<filenumber2>...]'
                    tt_filenums = moose.Mstring(tt_path+'/fileNumbers')
                    tt_filenums.value += '_' + filenums
            #### syn.Gbar remains the same, but we play with the weight which is a factor to Gbar
            #### The delay and weight can be set only after connecting a spike event generator.
            #### delay and weight are arrays: multiple event messages can be connected to a single synapse
            ## first argument below is the array index, we connect to the latest synapse created above
            ## But KinSynChan ignores weight of the synapse, so set the Gbar for it
            if libsyn.className == 'KinSynChan':
                syn.Gbar = weight*syn.Gbar
            else:
                ## note that you need to use Synapse (auto-created) under SynHandler
                ## to get/set weights , addSpike-s etc.
                ## wrap Synpase element by moose.Synapse(synhandler.path+'/synapse') or synhandler.synapse
                synhandler.synapse[-1].weight = weight
            synhandler.synapse[-1].delay = delay # seconds
            #print 'len = ',len(synhandler.synapse)
            #for i,syn_syn in enumerate(synhandler.synapse):
            #    print i,'th weight =',syn_syn.weight,'\n'



"""helper.py: 

Some helper functions which are compatible with both python2 and python3.
"""
    
__author__           = "Dilawar Singh"
__copyright__        = "Copyright 2017-, Dilawar Singh"
__version__          = "1.0.0"
__maintainer__       = "Dilawar Singh"
__email__            = "dilawars@ncbs.res.in"
__status__           = "Development"

import os
import re
import subprocess

def execute(cmd):
    """execute: Execute a given command.

    :param cmd: string, given command.

    Return:
    ------
        Return a iterator over output.
    """
    popen = subprocess.Popen(cmd, stdout=subprocess.PIPE, universal_newlines=True)
    for stdout_line in iter(popen.stdout.readline, ""):
        yield stdout_line 
    popen.stdout.close()
    return_code = popen.wait()
    if return_code:
        raise subprocess.CalledProcessError(return_code, cmd)


def find_files( dirname, ext=None, name_contains=None, text_regex_search=None):
    files = []
    for d, sd, fs in os.walk(dirname):
        for f in fs:
            fpath = os.path.join(d,f)
            include = True
            if ext is not None:
                if f.split('.')[-1] != ext:
                    include = False
            if name_contains:
                if name_contains not in os.path.basename(f):
                    include = False
            if text_regex_search:
                with open(fpath, 'r' ) as f:
                    txt = f.read()
                    if re.search(text_regex_search, txt) is None:
                        include = False
            if include:
                files.append(fpath)
    return files

# Matplotlib text for running simulation. It make sures at each figure is saved
# to individual png files.
matplotlibText = """
print( '>>>> saving all figues')
import matplotlib.pyplot as plt
def multipage(filename, figs=None, dpi=200):
    pp = PdfPages(filename)
    if figs is None:
        figs = [plt.figure(n) for n in plt.get_fignums()]
    for fig in figs:
        fig.savefig(pp, format='pdf')
    pp.close()

def saveall(prefix='results', figs=None):
    if figs is None:
        figs = [plt.figure(n) for n in plt.get_fignums()]
    for i, fig in enumerate(figs):
        outfile = '%s.%d.png' % (prefix, i)
        fig.savefig(outfile)
        print( '>>>> %s saved.' % outfile )
    plt.close()

try:
    saveall()
except Exception as e:
    print( '>>>> Error in saving: %s' % e )
    quit(0)
"""



#########################################################################
## This program is part of 'MOOSE', the
## Messaging Object Oriented Simulation Environment.
##           Copyright (C) 2014 Upinder S. Bhalla. and NCBS
## It is made available under the terms of the
## GNU Lesser General Public License version 2.1
## See the file COPYING.LIB for the full notice.
## Here test ConcChan. We set up a channel and a pump and run two tests.
## First we have zero flux through the pump and check that the conc
## equalizes on both sides of the membrane. Then we change the Kf of
## the pump to a value designed to give twice the conc In as Out.
## The system is a 1-voxel cyl compt and an internal endoMesh, having a
## volume = 1/8 of the cylinder.
##                      \
##    COMPT:    s ----> / s  :ENDO
##                 s =chan= s
##                      \
##
## This becomes:
##                                  \
##    COMPT:    s ----> s_xfer_endo / s     :ENDO
##                             s =chan= s
##                                  \
##
#########################################################################

import math

import numpy as np
import moose
import moose.fixXreacs as fixXreacs

print(moose.about())


def makeModel():
    # create container for model
    num = 1  # number of compartments
    model = moose.Neutral('/model')
    compartment = moose.CylMesh('/model/compartment')
    compartment.x1 = 1.0e-6  # Set it to a 1 micron single-voxel cylinder

    # create molecules and reactions
    s = moose.Pool('/model/compartment/s')
    rXfer = moose.Reac('/model/compartment/rXfer')
    #####################################################################
    # Put in endo compartment. Add molecule s
    endo = moose.EndoMesh('/model/endo')
    # Note that the chanPool must be on the 'In' compartment.
    #chanPool = moose.Pool( '/model/compartment/chanPool' )
    #chan = moose.ConcChan( '/model/compartment/chanPool/chan' )
    chanPool = moose.Pool('/model/endo/chanPool')
    chan = moose.ConcChan('/model/endo/chanPool/chan')
    endo.isMembraneBound = True
    endo.surround = compartment
    es = moose.Pool('/model/endo/s')
    #####################################################################
    moose.connect(rXfer, 'sub', s, 'reac')
    moose.connect(rXfer, 'prd', es, 'reac')
    moose.connect(chanPool, 'nOut', chan, 'setNumChan')
    moose.connect(chan, 'out', s, 'reac')
    moose.connect(chan, 'in', es, 'reac')
    volRatio = compartment.volume / endo.volume
    rXfer.Kf = 0.0  # 0.02/sec
    rXfer.Kb = 0.0  #
    s.concInit = 0.001
    chanPool.nInit = 1000.0
    # Flux (mM/s) = permeability * N * (conc_out - conc_in )
    chan.permeability = 0.1 * chanPool.volume * 6.022e23 / chanPool.nInit

    #####################################################################
    fixXreacs.fixXreacs('/model')
    #fixXreacs.restoreXreacs( '/model' )
    #fixXreacs.fixXreacs( '/model' )
    #####################################################################

    # Make solvers
    ksolve = moose.Ksolve('/model/compartment/ksolve')
    dsolve = moose.Dsolve('/model/dsolve')
    eksolve = moose.Ksolve('/model/endo/ksolve')
    edsolve = moose.Dsolve('/model/endo/dsolve')

    stoich = moose.Stoich('/model/compartment/stoich')
    stoich.compartment = compartment
    stoich.ksolve = ksolve
    stoich.dsolve = dsolve
    stoich.path = "/model/compartment/##"
    assert (dsolve.numPools == 2)

    estoich = moose.Stoich('/model/endo/stoich')
    estoich.compartment = endo
    estoich.ksolve = eksolve
    estoich.dsolve = edsolve
    estoich.path = "/model/endo/##"
    assert (edsolve.numPools == 2)

    edsolve.buildMeshJunctions(dsolve)

    plot1 = moose.Table2('/model/plot1')
    plot2 = moose.Table2('/model/plot2')
    plot3 = moose.Table2('/model/plot3')
    moose.connect('/model/plot1', 'requestOut', s, 'getN')
    moose.connect('/model/plot2', 'requestOut', es, 'getN')
    moose.connect('/model/plot3', 'requestOut',
                  '/model/compartment/s_xfer_endo', 'getN')
    plot4 = moose.Table2('/model/plot4')
    plot5 = moose.Table2('/model/plot5')
    plot6 = moose.Table2('/model/plot6')
    moose.connect('/model/plot4', 'requestOut', s, 'getConc')
    moose.connect('/model/plot5', 'requestOut', es, 'getConc')
    moose.connect('/model/plot6', 'requestOut',
                  '/model/compartment/s_xfer_endo', 'getConc')


def test_xchan(standalone=False):
    runtime = 100
    displayInterval = 2
    makeModel()
    for i in range(10, 18):
        moose.setClock(i, 0.01)
    moose.reinit()
    moose.start(runtime)
    s = moose.element('/model/compartment/s')
    es = moose.element('/model/endo/s')
    assert np.allclose(s.conc, es.conc)
    # We go for concEndo = 2x concOut. Then
    # We already know volIn = volOut/8
    # #in/volIn = 2x #out/volOut
    # #in/volIn = 2x #out/(8*volIn)
    # so #in = #out/4
    # From consv, #in + #out = nInit
    # so 5/4#out = nInit =>
    #       #out = 0.8*nInit; #in = 0.2*nInit
    #
    # flux = perm * nChan * (0.8nInit/volOut - 0.2nInit/(volOut/8) ) =
    #        perm * (nChan*nInit/volOut) * (0.8 - 0.2*8) =
    #        perm * (nChan*nInit/volOut) * (-0.8)
    #
    # This has to be balanced by flux = numKf * nOut = numKf * 0.8 * nInit
    # So 0.8*numKf*nInit = 0.8*perm*nChan*nInit/volOut

    # Note that chan.permeability = 0.1*chanPool.volume / chanPool.nInit

    # so numKf = perm*nChan/volOut = (0.1*volIn/nChan)*nChan/volOut=0.1/8
    #
    rXfer = moose.element('/model/compartment/rXfer')
    rXfer.numKf = 0.1 / 8.0
    moose.start(runtime)
    assert np.allclose(2 * s.conc, es.conc, atol=1e-5), (2*s.conc, es.conc)
    moose.delete('/model')


# Run the 'main' if this script is executed standalone.
if __name__ == '__main__':
    test_xchan(standalone=True)



# -*- coding: utf-8 -*-
from __future__ import print_function, division, absolute_import

# This is moose.server.
# It accepts simulation request on a specified TCP port (default 31417).
# It simulates the given file (usually a archive file e.g., tar.bz2) and sends
# back artefacts generated by simulation (mostly images); and streams data from
# moose.Tables back to client.

__author__           = "Dilawar Singh"
__copyright__        = "Copyright 2019, Dilawar Singh"
__version__          = "1.0.0"
__maintainer__       = "Dilawar Singh"
__email__            = "dilawars@ncbs.res.in"
__status__           = "Development"

import sys
import re
import os 
import time
import math
import shutil
import socket 
import signal
import tarfile 
import tempfile 
import threading 
import subprocess

import logging
logger_ = logging.getLogger('moose.server')

__all__ = [ 'serve' ]

# Global variable to stop all running threads.
stop_all_ = False
sock_     = None
stop_streamer_ = {}

# Use prefixL_ bytes to encode the size of stream. One can probably use just one
# byte to do. Lets go with the inefficient one for now.
prefixL_  = 9

# Matplotlib text for running simulation. It make sures at each figure is saved
# to individual png files.
matplotlibText = """
print( '>>>> saving all figues')
import matplotlib.pyplot as plt
def multipage(filename, figs=None, dpi=200):
    pp = PdfPages(filename)
    if figs is None:
        figs = [plt.figure(n) for n in plt.get_fignums()]
    for fig in figs:
        fig.savefig(pp, format='pdf')
    pp.close()

def saveall(prefix='results', figs=None):
    if figs is None:
        figs = [plt.figure(n) for n in plt.get_fignums()]
    for i, fig in enumerate(figs):
        outfile = '%s.%d.png' % (prefix, i)
        fig.savefig(outfile)
        print( '>>>> %s saved.' % outfile )
    plt.close()

try:
    saveall()
except Exception as e:
    print( '>>>> Error in saving: %s' % e )
    quit(0)
"""


def execute(cmd):
    """execute: Execute a given command.

    :param cmd: string, given command.

    Return:
    ------
        Return a iterator over output.
    """
    popen = subprocess.Popen(cmd, stdout=subprocess.PIPE, universal_newlines=True)
    for stdout_line in iter(popen.stdout.readline, ""):
        yield stdout_line 
    popen.stdout.close()
    return_code = popen.wait()
    if return_code:
        raise subprocess.CalledProcessError(return_code, cmd)


def find_files( dirname, ext=None, name_contains=None, text_regex_search=None):
    files = []
    for d, sd, fs in os.walk(dirname):
        for f in fs:
            fpath = os.path.join(d,f)
            include = True
            if ext is not None:
                if f.split('.')[-1] != ext:
                    include = False
            if name_contains:
                if name_contains not in os.path.basename(f):
                    include = False
            if text_regex_search:
                with open(fpath, 'r' ) as f:
                    txt = f.read()
                    if re.search(text_regex_search, txt) is None:
                        include = False
            if include:
                files.append(fpath)
    return files

def prefix_data_with_size(data):
    global prefixL_
    prefix = b'0'*(prefixL_-int(math.log10(len(data)))-1) + b'%d' % len(data)
    assert len(prefix) == prefixL_
    return b'%s%s' % (prefix, data)

# Signal handler.
def signal_handler(signum, frame):
    global stop_all_
    global sock_
    logger_.info( "User terminated all processes." )
    stop_all_ = True
    #  sock_.shutdown( socket.SHUT_RDWR )
    sock_.close()
    time.sleep(1)
    quit(1)


def split_data( data ):
    global prefixL_
    return data[:prefixL_].strip(), data[prefixL_:]

def send_msg(msg, conn, prefix='LOG'):
    if not msg.strip():
        return False
    if prefix != 'TAB':
        logger_.debug(msg)
    else:
        logger_.debug( 'Sending msg with size %d' % len(msg))
    msg = '<%s>%s' % (prefix, msg)
    conn.sendall(prefix_data_with_size(msg))

def run(cmd, conn, cwd=None):
    logger_.info( "Executing %s" % cmd )
    oldCWD = os.getcwd()
    if cwd is not None:
        os.chdir(cwd)
    try:
        for line in execute(cmd.split()):
            if line:
                send_msg(line, conn)
    except Exception as e:
        send_msg("Simulation failed: %s" % e, conn)
    os.chdir(oldCWD)

def recv_input(conn, size=1024):
    # first 10 bytes always tell how much to read next. Make sure the submit job
    # script has it
    d = conn.recv(prefixL_, socket.MSG_WAITALL)
    while len(d) < prefixL_:
        try:
            d = conn.recv(prefixL_, socket.MSG_WAITALL)
        except Exception:
            logger_.error("MSG FORMAT: %d bytes are size of msg."%prefixL_)
            continue
    d, data = int(d), b''
    while len(data) < d:
        data += conn.recv(d-len(data), socket.MSG_WAITALL)
    return data

def writeTarfile( data ):
    tfile = os.path.join(tempfile.mkdtemp(), 'data.tar.bz2')
    with open(tfile, 'wb' ) as f:
        logger_.info( "Writing %d bytes to %s" % (len(data), tfile))
        f.write(data)
    # Sleep for some time so that file can be written to disk.
    time.sleep(0.1)
    if not tarfile.is_tarfile(tfile):
        logger_.warning( 'Not a valid tar file: %s' % tfile)
        return None
    return tfile

def suffixMatplotlibStmt( filename ):
    outfile = '%s.1.py' % filename
    with open(filename, 'r') as f:
        txt = f.read()

    with open(outfile, 'w' ) as f:
        f.write( txt )
        f.write( '\n' )
        f.write( matplotlibText )
    return outfile

def streamer_client(socketPath, conn):
    # Connect to running socket server.
    global stop_streamer_
    stop = False
    logger_.debug( "Trying to connect to server at : %s" % socketPath )
    while not os.path.exists( socketPath ):
        #print( 'socket %s is not available yet.' % socketPath )
        time.sleep(0.1)
        stop = stop_streamer_[threading.currentThread().name]
        if stop:
            return

    stClient = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
    try:
        stClient.connect(socketPath)
    except socket.error as e:
        logger_.warning('Could not connect: %s' % e)
        return

    # send streaming data back to client. The streamer send fixed size messages
    # of 1024/2048 bytes each (see the c++ implmenetation).
    logger_.info( "Socket Streamer is connected with server." )
    stClient.settimeout(0.05)
    send_msg( b'Now streaming table data.', conn, 'TAB')
    while not stop:
        stop = stop_streamer_[threading.currentThread().name]
        data = b''
        try:
            data = stClient.recv(1024)
            if len(data.strip()) > 0:
                send_msg(data, conn, 'TAB')
        except socket.timeout:
            continue
    stClient.close()
    if os.path.isfile(socketPath):
        os.unlink(socketPath)

def run_file(filename, conn, cwd=None):
    # set environment variable so that socket streamer can start.
    global stop_streamer_
    socketPath = os.path.join(tempfile.mkdtemp(), 'SOCK_TABLE_STREAMER')
    os.environ['MOOSE_STREAMER_ADDRESS'] = socketPath
    streamerThread = threading.Thread(target=streamer_client
            , args=(socketPath, conn,))
    stop_streamer_[streamerThread.name] = False
    streamerThread.daemon = True
    streamerThread.start()
    filename = suffixMatplotlibStmt(filename)
    run( "%s %s" % (sys.executable, filename), conn, cwd)
    stop_streamer_[streamerThread.name] = True
    streamerThread.join( timeout = 1)
    if streamerThread.is_alive():
        logger_.error( "The socket streamer client is still running...")

def extract_files(tfile, to):
    userFiles = []
    with tarfile.open(tfile, 'r' ) as f:
        userFiles = f.getnames( )
        try:
            f.extractall( to )
        except Exception as e:
            logger_.warning( e)
    # now check if all files have been extracted properly
    for f in userFiles:
        if not os.path.exists(f):
            logger_.error( "File %s could not be extracted." % f )
    return userFiles

def prepareMatplotlib( cwd ):
    with open(os.path.join(cwd, 'matplotlibrc'), 'w') as f:
        f.write( 'interactive : True' )

def send_bz2(conn, data):
    global prefixL_
    send_msg(data, conn, 'TAR')

def sendResults(tdir, conn, notTheseFiles):
    # Only send new files.
    resdir = tempfile.mkdtemp()
    resfile = os.path.join(resdir, 'results.tar.bz2')
    with tarfile.open( resfile, 'w|bz2') as tf:
        for f in find_files(tdir, ext='png'):
            logger_.info( "Adding file %s" % f )
            tf.add(f, os.path.basename(f))

    time.sleep(0.01)
    # now send the tar file back to client
    with open(resfile, 'rb' ) as f:
        data = f.read()
        logger_.info( 'Total bytes to send to client: %d' % len(data))
        send_bz2(conn, data)
    shutil.rmtree(resdir)

def find_files_to_run( files ):
    """Any file name starting with __main is to be run.
    Many such files can be recieved by client.
    """
    toRun = []
    for f in files:
        if '__main' in os.path.basename(f):
            toRun.append(f)
    if toRun:
        return toRun
    # Else guess.
    if len(files) == 1:
        return files

    for f in files:
        with open(f, 'r' ) as fh:
            txt = fh.read()
            if re.search(r'def\s+main\(', txt):
                if re.search(r'^\s+main\(\S+?\)', txt):
                    toRun.append(f)
    return toRun

def simulate( tfile, conn ):
    """Simulate a given tar file.
    """
    tdir = os.path.dirname( tfile )
    os.chdir( tdir )
    userFiles = extract_files(tfile, tdir)
    # Now simulate.
    toRun = find_files_to_run(userFiles)
    if len(toRun) < 1:
        return 1
    prepareMatplotlib(tdir)
    status, msg = 0, ''
    for _file in toRun:
        try:
            run_file(_file, conn, tdir) 
        except Exception as e:
            msg += str(e)
            status = 1
    return status, msg

def savePayload( conn ):
    data = recv_input(conn)
    tarfileName = writeTarfile(data)
    return tarfileName, len(data)

def handle_client(conn, ip, port):
    isActive = True
    logger_.info( "Serving request from %s:%s" % (ip, port) )
    while isActive:
        tarfileName, nBytes = savePayload(conn)
        if tarfileName is None:
            logger_.warning( "Could not recieve data." )
            isActive = False
        if not os.path.isfile(tarfileName):
            send_msg("[ERROR] %s is not a valid tarfile. Retry"%tarfileName, conn)
            break

        # list of files before the simulation.
        notthesefiles = find_files(os.path.dirname(tarfileName))
        res, msg = simulate( tarfileName, conn )
        if 0 != res:
            send_msg( "Failed to run simulation: %s" % msg, conn)
            isActive = False
            time.sleep(0.1)

        # Send results after DONE is sent.
        send_msg('All done', conn, 'EOS')
        sendResults(os.path.dirname(tarfileName), conn, notthesefiles)
        break


def start_server( host, port, max_requests = 10 ):
    global stop_all_
    global sock_
    sock_ = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    sock_.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
    try:
        sock_.bind( (host, port))
        logger_.info( "Server created %s:%s" %(host,port) )
    except Exception as e:
        logger_.error( "Failed to bind: %s" % e)
        quit(1)

    # listen upto 10 of requests
    sock_.listen(max_requests)
    while True:
        if stop_all_:
            break
        sock_.settimeout(10)
        try:
            conn, (ip, port) = sock_.accept()
        except socket.timeout as e:
            continue
        sock_.settimeout(0.0)
        t = threading.Thread(target=handle_client, args=(conn, ip, port)) 
        t.start()
    sock_.close()

def serve(host, port):
    start_server(host, port)

def main( args ):
    global stop_all_
    host, port = args.host, args.port
    # Install a signal handler.
    signal.signal( signal.SIGINT, signal_handler)
    serve(host, port)

if __name__ == '__main__':
    import argparse
    # Argument parser.
    description = '''Run MOOSE server.'''
    parser = argparse.ArgumentParser(description=description, add_help=False)
    parser.add_argument( '--help', action='help', help='Show this msg and exit')
    parser.add_argument('--host', '-h'
        , required = False, default = socket.gethostbyname(socket.gethostname())
        , help = 'Server Name'
        )
    parser.add_argument('--port', '-p'
        , required = False, default = 31417, type=int
        , help = 'Port number'
        )
    class Args: pass 
    args = Args()
    parser.parse_args(namespace=args)
    try:
        main(args)
    except KeyboardInterrupt as e:
        stop_all_ = True
        quit(1)



# -*- coding: utf-8 -*-
#
#   Copyright (c) 2007, Simon Edwards <simon@simonzone.com>
#    All rights reserved.
#
#    Redistribution and use in source and binary forms, with or without
#    modification, are permitted provided that the following conditions are met:
#        * Redistributions of source code must retain the above copyright
#        notice, this list of conditions and the following disclaimer.
#        * Redistributions in binary form must reproduce the above copyright
#        notice, this list of conditions and the following disclaimer in the
#        documentation and/or other materials provided with the distribution.
#        * Neither the name of the  Simon Edwards <simon@simonzone.com> nor the
#        names of its contributors may be used to endorse or promote products
#        derived from this software without specific prior written permission.
#
#    THIS SOFTWARE IS PROVIDED BY Simon Edwards <simon@simonzone.com> ''AS IS'' AND ANY
#    EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
#    WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#    DISCLAIMED. IN NO EVENT SHALL Simon Edwards <simon@simonzone.com> BE LIABLE FOR ANY
#    DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
#    (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
#    LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
#    ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
#    (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
#    SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# FindPyQt.py
# Copyright (c) 2007, Simon Edwards <simon@simonzone.com>
# Redistribution and use is allowed according to the terms of the BSD license.
# For details see the accompanying COPYING-CMAKE-SCRIPTS file.

try:
    import PyQt4.pyqtconfig
    pyqtcfg = PyQt4.pyqtconfig.Configuration()
    pyqt_sip_flags = pyqtcfg.pyqt_sip_flags
    print("pyqt_version:%06.0x" % pyqtcfg.pyqt_version)
    print("pyqt_version_num:%d" % pyqtcfg.pyqt_version)
    print("pyqt_version_str:%s" % pyqtcfg.pyqt_version_str)
except ImportError:
    # PyQt built with configure-ng.py has no pyqtconfig module
    from PyQt4.QtCore import PYQT_CONFIGURATION, PYQT_VERSION, PYQT_VERSION_STR
    pyqtcfg = None
    pyqt_sip_flags = PYQT_CONFIGURATION["sip_flags"]
    print("pyqt_version:%06.0x" % PYQT_VERSION)
    print("pyqt_version_num:%d" % PYQT_VERSION)
    print("pyqt_version_str:%s" % PYQT_VERSION_STR)

pyqt_version_tag = ""
in_t = False
for item in pyqt_sip_flags.split(' '):
    if item=="-t":
        in_t = True
    elif in_t:
        if item.startswith("Qt_4"):
            pyqt_version_tag = item
    else:
        in_t = False
print("pyqt_version_tag:%s" % pyqt_version_tag)

if pyqtcfg:
    print("pyqt_mod_dir:%s" % pyqtcfg.pyqt_mod_dir)
    print("pyqt_sip_dir:%s" % pyqtcfg.pyqt_sip_dir)
    print("pyqt_sip_flags:%s" % pyqtcfg.pyqt_sip_flags)
    print("pyqt_bin_dir:%s" % pyqtcfg.pyqt_bin_dir)
else:
    import sipconfig
    sipcfg = sipconfig.Configuration()
    print("pyqt_mod_dir:%s" % sipcfg.default_mod_dir)
    print("pyqt_sip_dir:%s" % sipcfg.default_sip_dir)
    print("pyqt_sip_flags:%s" % pyqt_sip_flags)
    print("pyqt_bin_dir:%s" % sipcfg.default_bin_dir)






# -*- coding: utf-8 -*-
# test_reader.py ---
#
# Filename: test_reader.py
# Description:
# Author:
# Maintainer:
# Created: Wed Jul 24 16:02:21 2013 (+0530)
# Version:
# Last-Updated: Sun Apr 17 16:13:01 2016 (-0400)
#           By: subha
#     Update #: 112
# URL:
# Keywords:
# Compatibility:
#
#

# Commentary:
#
#
#
#

# Change log:
#
#
#
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 3, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; see the file COPYING.  If not, write to
# the Free Software Foundation, Inc., 51 Franklin Street, Fifth
# Floor, Boston, MA 02110-1301, USA.
#
#

# Code:

from __future__ import print_function
import unittest
import numpy as np
import moose
import neuroml as nml
from reader import NML2Reader

class TestFullCell(unittest.TestCase):
    def setUp(self):
        self.reader = NML2Reader(verbose=True)

        self.lib = moose.Neutral('/library')
        self.filename = 'test_files/NML2_FullCell.nml'
        self.reader.read(self.filename)
        for ncell in self.reader.nml_to_moose:
            if isinstance(ncell, nml.Cell):
                self.ncell = ncell
                break
        self.mcell = moose.element('/library/SpikingCell')
        self.soma = moose.element(self.mcell.path + '/Soma')
        self.dendrite1 = moose.element(self.mcell.path + '/Dendrite1')
        self.dendrite2 = moose.element(self.mcell.path + '/Dendrite2')
        self.spine1 = moose.element(self.mcell.path + '/Spine1')

    def test_basicLoading(self):
        self.assertEqual(self.reader.filename, self.filename, 'filename was not set')
        self.assertIsNotNone(self.reader.doc, 'doc is None')

    def test_createCellPrototype(self):
        self.assertIsInstance(self.mcell, moose.Neuron)
        self.assertEqual(self.mcell.name, self.ncell.id)

    def test_createMorphology(self):
        for comp_id in moose.wildcardFind(self.mcell.path + '/##[ISA=Compartment]'):
            comp = moose.element(comp_id)
            p0 = self.reader.moose_to_nml[comp].proximal
            if p0:
                self.assertAlmostEqual(comp.x0, float(p0.x)*1e-6) # Assume micron unit for length
                self.assertAlmostEqual(comp.y0, float(p0.y)*1e-6)
                self.assertAlmostEqual(comp.z0, float(p0.z)*1e-6)
            p1 = self.reader.moose_to_nml[comp].distal
            self.assertAlmostEqual(comp.x, float(p1.x)*1e-6)
            self.assertAlmostEqual(comp.y, float(p1.y)*1e-6)
            self.assertAlmostEqual(comp.z, float(p1.z)*1e-6)

    def test_connectivity(self):
        """Test raxial-axial connectivity between MOOSE compartments when
        there is parent->child relation in NML2."""
        id_to_seg = dict([(seg.id, seg) for seg in self.ncell.morphology.segments])
        for seg in self.ncell.morphology.segments:
            try:
                pseg = id_to_seg[seg.parent.segment]
            except AttributeError:
                continue
            comp = self.reader.nml_to_moose[seg]
            pcomp = self.reader.nml_to_moose[pseg]
            
            '''
            TODO: what should this be updated to???
            self.assertIn(comp.id_, pcomp.neighbours['raxial'])'''

    def test_capacitance(self):
        for comp_id in moose.wildcardFind(self.mcell.path + '/##[ISA=Compartment]'):
            comp = moose.element(comp_id)
            # We know that a few um^2 compartment with uF/cm^2 specific capacitance must be around a pico Farad.
            self.assertTrue((comp.Cm > 0) and (comp.Cm < 1e-6))

    def test_protochans(self):
        """TODO: verify the prototype cahnnel."""
        for chan_id in moose.wildcardFind('/library/##[ISA=HHChannel]'):
            print(moose.element(chan_id))

    def test_HHChannels(self):
        """Verify copied channel in membrane properties."""
        self.assertTrue(moose.exists(self.soma.path + '/naChansSoma'))
        soma_na = moose.element(self.soma.path+'/naChansSoma')
        chans = moose.wildcardFind(self.mcell.path + '/##[ISA=HHChannel]')
        self.assertTrue(len(chans) < 3) # Only soma and dendrite2 have the channels
        self.assertAlmostEqual(soma_na.Gbar, 120e-2 * self.soma.diameter * self.soma.diameter * np.pi, places=6)

'''
Not yet working in NML2...

class TestGran98(unittest.TestCase):
    def setUp(self):
        self.reader = NML2Reader()
        self.lib = moose.Neutral('/library')
        self.filename = 'test_files/Granule_98/Granule_98.nml'
        self.reader.read(self.filename)
        for ncell in self.reader.nml_to_moose:
            if isinstance(ncell, nml.Cell):
                self.ncell = ncell
                break
        self.mcell = moose.element(moose.wildcardFind('/##[ISA=Cell]')[0])

    def test_CaPool(self):
        pass
'''

if __name__ == '__main__':
    unittest.main()

#
# test_reader.py ends here



# -*- coding: utf-8 -*-
from __future__ import print_function, division, absolute_import

####################################################################
# fixXreacs.py
# The program is meant to take a model and reconfigure any cross-compartment
# reactions so that they are split into portions on either side coupled
# either by diffusion, or by a reaction-driven translocation process.
#
# Copyright (C) Upinder S. Bhalla       2018
# This program is free software. It is licensed under the terms of
# GPL version 3 or later at your discretion.
# This program carries no warranty whatsoever.
####################################################################

import moose._moose as _moose
import moose

msgSeparator = "_xMsg_"

def findCompt(elm):
    elm = moose.element(elm)
    pa = elm.parent
    while pa.path != '/':
        if moose.Neutral(pa).isA['ChemCompt']:
            return pa.path
        pa = pa.parent
    print('Error: No compartment parent found for ' + elm.path)
    return '/'


# http://stackoverflow.com/q/3844948/
def checkEqual(lst):
    return not lst or lst.count(lst[0]) == len(lst)


def findXreacs(basepath, reacType):
    reacs = _moose.wildcardFind(basepath + '/##[ISA=' + reacType + 'Base]')
    ret = []
    for i in reacs:
        reacc = findCompt(i)
        subs = i.neighbors['subOut']
        prds = i.neighbors['prdOut']
        subc = [findCompt(j) for j in subs]
        prdc = [findCompt(j) for j in prds]

        enzc = []
        if reacType == 'Enz':
            enzc = [reacc]
        if not checkEqual(subc + prdc + enzc):
            ret.append([i, reacc, subs, subc, prds, prdc])
    return ret


def removeEnzFromPool(pool):
    kids = _moose.wildcardFind(pool.path + "/#")
    for i in kids:
        if i.isA['EnzBase']:
            _moose.delete(i)
        elif i.isA['Function']:
            _moose.delete(i)


# If a pool is not in the same compt as reac, make a proxy in the reac
# compt, connect it up, and disconnect the one in the old compt.
def proxify(reac, reacc, direction, pool, poolc):
    reacc_elm = moose.element(reacc)
    reac_elm = moose.element(reac)
    # Preserve the rates which were set up for the x-compt reacn
    #_moose.showfield( reac )
    dupname = pool.name + '_xfer_' + moose.element(poolc).name
    #  print("#############", pool, dupname, poolc)
    if _moose.exists(reacc + '/' + dupname):
        duppool = moose.element(reacc + '/' + dupname)
    else:
        # This also deals with cases where the duppool is buffered.
        duppool = _moose.copy(pool, reacc_elm, dupname)
    duppool.diffConst = 0  # diffusion only happens in original compt
    removeEnzFromPool(duppool)
    disconnectReactant(reac, pool, duppool)
    moose.connect(reac, direction, duppool, 'reac')
    #_moose.showfield( reac )
    #_moose.showmsg( reac )


def enzProxify(enz, enzc, direction, pool, poolc):
    if enzc == poolc:
        return
    enze = moose.element(enz)
    # kcat and k2 are indept of volume, just time^-1
    km = enze.numKm
    proxify(enz, enzc, direction, pool, poolc)
    enze.numKm = km


def reacProxify(reac, reacc, direction, pool, poolc):
    if reacc == poolc:
        return
    reac_elm = moose.element(reac)
    kf = reac_elm.numKf
    kb = reac_elm.numKb
    proxify(reac, reacc, direction, pool, poolc)
    reac_elm.numKf = kf
    reac_elm.numKb = kb


def identifyMsg(src, srcOut, dest):
    if src.isA['ReacBase'] or src.isA['EnzBase']:
        if srcOut == 'subOut':
            return msgSeparator + src.path + ' sub ' + dest.path + ' reac'
        if srcOut == 'prdOut':
            return msgSeparator + src.path + ' prd ' + dest.path + ' reac'
    return ''


def disconnectReactant(reacOrEnz, reactant, duppool):
    outMsgs = reacOrEnz.msgOut
    infoPath = duppool.path + '/info'
    if _moose.exists(infoPath):
        info = moose.element(infoPath)
    else:
        info = moose.Annotator(infoPath)

    #_moose.le( reactant )
    notes = ""
    #_moose.showmsg( reacOrEnz )
    for i in outMsgs:
        #print("killing msg from {} to {}\nfor {} and {}".format(
        #    reacOrEnz.path, reactant.path, i.srcFieldsOnE1,
        #    i.srcFieldsOnE2))
        if i.e1 == reactant:
            msgStr = identifyMsg(i.e2, i.e2.srcFieldsOnE2[0], i.e1)
            if len(msgStr) > 0:
                notes += msgStr
                _moose.delete(i)
        elif i.e2 == reactant:
            msgStr = identifyMsg(i.e1[0], i.srcFieldsOnE1[0], i.e2[0])
            if len(msgStr) > 0:
                notes += msgStr
                _moose.delete(i)
    #  print("MSGS to rebuild:", notes)
    info.notes += notes

def fixXreacs(basepath):
    xr = findXreacs(basepath, 'Reac')
    xe = findXreacs(basepath, 'Enz')

    for i in (xr):
        reac, reacc, subs, subc, prds, prdc = i
        for j in range(len(subs)):
            reacProxify(reac, reacc, 'sub', subs[j], subc[j])
        for j in range(len(prds)):
            reacProxify(reac, reacc, 'prd', prds[j], prdc[j])

    for i in (xe):
        reac, reacc, subs, subc, prds, prdc = i
        for j in range(len(subs)):
            enzProxify(reac, reacc, 'sub', subs[j], subc[j])
        for j in range(len(prds)):
            enzProxify(reac, reacc, 'prd', prds[j], prdc[j])


#####################################################################


def getOldRates(msgs):
    if len(msgs) > 1:
        m1 = msgs[1].split(msgSeparator)[0]
        elm = moose.element(m1.split(' ')[0])
        if elm.isA['ReacBase']:
            return [elm.numKf, elm.numKb]
        elif elm.isA['EnzBase']:
            return [
                elm.numKm,
            ]
    print("Warning: getOldRates did not have any messages")
    return [
        0,
    ]


def restoreOldRates(oldRates, msgs):
    #print oldRates, msgs
    if len(msgs) > 1:
        m1 = msgs[1].split(msgSeparator)[0]
        elm = moose.element(m1.split(' ')[0])
        if elm.isA['ReacBase']:
            elm.numKf = oldRates[0]
            elm.numKb = oldRates[1]
        elif elm.isA['enzBase']:
            elm.numKm = oldRates[0]


def restoreXreacs(basepath):
    proxyInfo = _moose.wildcardFind(basepath + "/##/#_xfer_#/info")
    for i in proxyInfo:
        msgs = i.notes.split(msgSeparator)
        oldRates = getOldRates(msgs)
        #print( "Deleting {}".format( i.parent.path ) )
        _moose.delete(i.parent)
        for j in msgs[1:]:
            if len(j) > 0:
                args = j.split(' ')
                assert (len(args) == 4)
                #_moose.showfield( args[0] )
                #check to see if object exist before moose.connect, cases where object is deleted but
                #_xref_ annotation info field still holds the reference
                if (_moose.exists(args[0]) and _moose.exists(args[2])):
                    moose.connect(args[0], args[1], args[2], args[3])
                #print( "Reconnecting {}".format( args ) )
                #_moose.showfield( args[0] )
        restoreOldRates(oldRates, msgs)



# -*- coding: utf-8 -*-
# Issue 47 on moose-core
import moose
import os
sdir = os.path.dirname( __file__ )
print( '[INFO] Using moose from %s' % moose.__file__ )
print( '[WONTFIX] See https://github.com/BhallaLab/moose-core/issues/47')
if False:
    moose.loadModel(os.path.join(sdir,'../data/acc94.g'),'/acc94','gsl')
    #moose.loadModel(os.path.join(sdir,'../data/acc94.g'),'/acc94','gssa')
    moose.reinit()
    moose.start( 10 )
else:
    quit(0)



# -*- coding: utf-8 -*-
"""params.py

Parameters used in this model

These parameters are from paper Miller et. al. "The stability of CaMKII
switch"

"""

__author__           = "Dilawar Singh"
__copyright__        = "Copyright 2015, Dilawar Singh and NCBS Bangalore"
__credits__          = ["NCBS Bangalore"]
__license__          = "GNU GPL"
__version__          = "1.0.0"
__maintainer__       = "Dilawar Singh"
__email__            = "dilawars@ncbs.res.in"
__status__           = "Development"


run_time             = 30
N_CaMK               = 10
N_PP1                = 100
num_switch           = 1
voxel_length         = 125e-9
num_voxels           = 1

diff_consts          = { 'x' : 1e-13, 'y' : 1e-13, 'pp1' : 1e-13 }

conc_i1_free         = 0.1e-3
act_CaN              = 1.0
act_PKA              = 1.0

# Michaelis constant of protein phosphatase.
# 0.4 um to 2.0 uM have been used. Miller uses 0.2 um. The switch shows
# bistability of these ranges. We have taken the largest Km (or slowest) first
# step in dephosphorylation.
K_M                  = 10e-3
k_2                  = 10.0

# Hill coefficientfor Ca++ activation of CaMKII
K_H1                 = 0.7e-3
K_H2                 = 0.3e-3

k_1                  = 1.5
k_3                  = 100e3
k_4                  = 0.001
K_I                  = 1e-6

rate_loosex          = 0.1
rate_loosey          = 0.1
rate_gainx           = 1
rate_gainy           = 1

turnover_rate        = 1/(30*3600.0)
v_1                  = 1.268e-5
v_2                  = 4.36e-3
phosphatase_inhibit  = 280.0
vi                   = phosphatase_inhibit

## Calcium input expression.
ca_basal             = 80e-6
ca_expr              = "(fmod(t,4)<2)?{0}:({0}*(1+0.5*rand(-1)))".format( ca_basal )



# -*- coding: utf-8 -*-
from __future__ import print_function, division

__author__           = "Dilawar Singh"
__copyright__        = "Copyright 2016, Dilawar Singh"
__credits__          = ["NCBS Bangalore"]
__license__          = "GNU GPL"
__version__          = "1.0.0"
__maintainer__       = "Dilawar Singh"
__email__            = "dilawars@ncbs.res.in"
__status__           = "Development"

import os
import sys
sys.path.append(os.path.dirname(__file__))
import time
import socket
import numpy as np
import multiprocessing as mp
import moose
import moose.utils as mu
import models
from collections import defaultdict

print( '[INFO] Using moose form %s' % moose.__file__ )

port_ = 3114

def socket_client(q, done):
    global port_
    # This function waits for socket to be available.
    s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)

    while 1:
        if 1 == done.value:
            print( '[INFO] MOOSE is done before I could connect' )
            q.put({})
            return
        s.settimeout(0.5)
        try:
            s.connect( ('127.0.0.1', port_) )
        except socket.timeout as e:
            print(e)
            break

    print( 'Py: Connected' )

    # This is client reponsibility to read the data.
    s.settimeout(0.01)
    data = b''
    while True:
        try:
            data += s.recv(1024)
        except socket.timeout:
            print(end='x')
        if 1 == done.value:
            break
    if data:
        res = mu.decode_data(data)
        q.put(res)
    else:
        q.put({})
    s.close()
    return

def test():
    global finish_all_
    os.environ['MOOSE_STREAMER_ADDRESS'] = 'http://127.0.0.1:%d'%port_
    done = mp.Value('d', 0)
    q = mp.Queue()
    client = mp.Process(target=socket_client, args=(q, done))
    client.start()

    time.sleep(0.1)

    print( '[INFO] Socket client is running now' )
    ts = models.simple_model_a()
    moose.reinit()
    time.sleep(0.1)
    # If TCP socket is created, some delay is often neccessary before start. Don't
    # know why. probably some latency in a fresh TCP socket. A TCP guru can
    # tell.
    moose.start(50)
    print( 'MOOSE is done' )

    time.sleep(0.5)
    done.value = 1
    res = q.get()
    client.join()

    if not res:
        raise RuntimeWarning('Nothing was streamed')
    for k in res:
        a = res[k][1::2]
        b = moose.element(k).vector
        print(k, len(a), len(b))
        assert( (a==b).all())
    print( 'Test 1 passed' )

def main( ):
    test( )
    print( '[INFO] All tests passed' )

if __name__ == '__main__':
    main()



#########################################################################
## This program is part of 'MOOSE', the
## Messaging Object Oriented Simulation Environment.
##           Copyright (C) 2014 Upinder S. Bhalla. and NCBS
## It is made available under the terms of the
## GNU Lesser General Public License version 2.1
## See the file COPYING.LIB for the full notice.
#########################################################################

import os
import moose
print("[INFO ] using moose from %s" % moose.__file__)
import numpy as np 

sdir_ = os.path.dirname(os.path.realpath(__file__))


def runAndSavePlots( name ):
    runtime = 20.0
    moose.reinit()
    moose.start( runtime )
    pa = moose.Neutral( '/model/graphs/' + name )
    for x in moose.wildcardFind( '/model/#graphs/conc#/#' ):
        if ( x.tick != -1 ):
            tabname = '/model/graphs/' + name + '/' + x.name + '.' + name
            y = moose.Table( tabname )
            y.vector = x.vector
            y.tick = -1

# Takes args ee, gsl, or gssa
def switchSolvers( solver ):
        if ( moose.exists( 'model/kinetics/stoich' ) ):
            moose.delete( '/model/kinetics/stoich' )
            moose.delete( '/model/kinetics/ksolve' )
        compt = moose.element( '/model/kinetics' )
        if ( solver == 'gsl' ):
            ksolve = moose.Ksolve( '/model/kinetics/ksolve' )
        if ( solver == 'gssa' ):
            ksolve = moose.Gsolve( '/model/kinetics/ksolve' )
        if ( solver != 'ee' ):
            stoich = moose.Stoich( '/model/kinetics/stoich' )
            stoich.compartment = compt
            stoich.ksolve = ksolve
            stoich.path = "/model/kinetics/##"

def test_switch_solvers():
    """
    At zero order, you can select the solver you want to use within the
    function moose.loadModel( filename, modelpath, solver ).
    Having loaded in the model, you can change the solver to use on it.
    This example illustrates how to assign and change solvers for a
    kinetic model. This process is necessary in two situations:

    * If we want to change the numerical method employed, for example,
      from deterministic to stochastic.
    * If we are already using a solver, and we have changed the reaction
      network by adding or removing molecules or reactions.

    Note that we do not have to change the solvers if the volume or
    reaction rates change.
    In this example the model is loaded in with a gsl solver. The
    sequence of solver calculations is:

    #. gsl
    #. ee
    #. gsl
    #. gssa
    #. gsl

    If you're removing the solvers, you just delete the stoichiometry
    object and the associated ksolve/gsolve. Should there be diffusion
    (a dsolve)then you should delete that too. If you're
    building the solvers up again, then you must do the following
    steps in order:

    #. build up the ksolve/gsolve and stoich (any order)
    #. Assign stoich.ksolve
    #. Assign stoich.path.

    See the Reaction-diffusion section should you want to do diffusion
    as well.

    """

    solver = "gsl"  # Pick any of gsl, gssa, ee..
    mfile = os.path.join(sdir_, '../data/kkit_objects_example.g')
    modelId = moose.loadModel( mfile, 'model', solver )
    # Increase volume so that the stochastic solver gssa
    # gives an interesting output
    compt = moose.element( '/model/kinetics' )
    compt.volume = 1e-19
    runAndSavePlots( 'gsl' )
    #########################################################
    switchSolvers( 'ee' )
    runAndSavePlots( 'ee' )
    #########################################################
    switchSolvers( 'gsl' )
    runAndSavePlots( 'gsl2' )
    #########################################################
    switchSolvers( 'gssa' )
    runAndSavePlots( 'gssa' )
    #########################################################
    switchSolvers( 'gsl' )
    runAndSavePlots( 'gsl3' )
    #########################################################

    # Display all plots.
    plotdt = moose.element( '/clock' ).tickDt[18]
    conc = []
    for x in moose.wildcardFind( '/model/#graphs/conc#/#' ):
        conc.append(x.vector)
    conc = np.array(conc)
    assert conc.mean() > 0.0

    data = []
    for x in moose.wildcardFind( '/model/graphs/gsl/#' ):
        data.append(x.vector)
    gsl = np.array(data)
    assert abs(conc - gsl).sum() < 0.25

    data=[]
    for x in moose.wildcardFind( '/model/graphs/ee/#' ):
        data.append(x.vector)
    ee = np.array(data)
    assert abs(conc-ee).sum() < 0.2

    data=[]
    for x in moose.wildcardFind( '/model/graphs/gsl2/#' ):
        data.append(x.vector)
    gsl2 = np.array(data)
    assert abs(conc-gsl2).sum() == 0.0   # these are the same.

    data=[]
    for x in moose.wildcardFind( '/model/graphs/gssa/#' ):
        data.append(x.vector)
    gssa = np.array(data)
    assert abs(conc - gssa).sum() < 0.15, (conc - gssa).sum()
    assert gssa.shape == conc.shape == gsl.shape == ee.shape 

    print('all done')

def main():
    test_switch_solvers()

if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
# run_hhcell.py ---
#
# Filename: run_hhcell.py
# Description:
# Author:
# Maintainer: P Gleeson
# Version:
# URL:
# Keywords:
# Compatibility:
#
#

# Commentary:
#
#
#
#

# Change log:
#
#
#
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 3, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; see the file COPYING.  If not, write to
# the Free Software Foundation, Inc., 51 Franklin Street, Fifth
# Floor, Boston, MA 02110-1301, USA.
#
#

# Code:

import moose
import sys
from reader import NML2Reader
import numpy as np


def test_channel_gates():
    """Creates prototype channels under `/library` and plots the time
    constants (tau) and activation (minf, hinf, ninf) parameters for the
    channel gates.

    """
    import matplotlib.pyplot as plt
    lib = moose.Neutral('/library')
    m = moose.element('/library[0]/naChan[0]/gateX')
    h = moose.element('/library[0]/naChan[0]/gateY')
    n = moose.element('/library[0]/kChan[0]/gateX')
    v = np.linspace(n.min,n.max, n.divs+1)
    
    plt.subplot(221)
    plt.plot(v, 1/m.tableB, label='tau_m')
    plt.plot(v, 1/h.tableB, label='tau_h')
    plt.plot(v, 1/n.tableB, label='tau_n')
    plt.legend()
    
    plt.subplot(222)
    plt.plot(v, m.tableA/m.tableB, label='m_inf')
    plt.plot(v, h.tableA/h.tableB, label='h_inf')
    plt.plot(v, n.tableA/n.tableB, label='n_inf')
    plt.legend()
    
    plt.subplot(223)
    plt.plot(v, m.tableA, label='mA(alpha)')
    plt.plot(v, h.tableA, label='hA(alpha)')
    plt.plot(v, n.tableA, label='nA(alpha)')
    plt.legend()
    plt.subplot(224)
    
    plt.plot(v, m.tableB, label='mB')
    plt.plot(v, m.tableB-m.tableA, label='mB-A(beta)')
    
    plt.plot(v, h.tableB, label='hB')
    plt.plot(v, h.tableB-h.tableA, label='hB-A(beta)')
    
    plt.plot(v, n.tableB, label='nB')
    plt.plot(v, n.tableB-n.tableA, label='nB-nA(beta)')
    plt.legend()
    
    plt.show()


def run(nogui):
    
    reader = NML2Reader(verbose=True)

    filename = 'test_files/NML2_SingleCompHHCell.nml'
    print('Loading: %s'%filename)
    reader.read(filename, symmetric=True)
    
    
    msoma = reader.getComp(reader.doc.networks[0].populations[0].id,0,0)
    print(msoma)
    
    
    data = moose.Neutral('/data')
    
    pg = reader.getInput('pulseGen1')
    
    inj = moose.Table('%s/pulse' % (data.path))
    moose.connect(inj, 'requestOut', pg, 'getOutputValue')
    
    
    vm = moose.Table('%s/Vm' % (data.path))
    moose.connect(vm, 'requestOut', msoma, 'getVm')
    
    simdt = 1e-6
    plotdt = 1e-4
    simtime = 300e-3
    #moose.showmsg( '/clock' )
    for i in range(8):
        moose.setClock( i, simdt )
    moose.setClock( 8, plotdt )
    moose.reinit()
    moose.start(simtime)
    
    print("Finished simulation!")
    
    t = np.linspace(0, simtime, len(vm.vector))
    
    if not nogui:
        import matplotlib.pyplot as plt

        vfile = open('moose_v_hh.dat','w')

        for i in range(len(t)):
            vfile.write('%s\t%s\n'%(t[i],vm.vector[i]))
        vfile.close()
        plt.subplot(211)
        plt.plot(t, vm.vector * 1e3, label='Vm (mV)')
        plt.legend()
        plt.title('Vm')
        plt.subplot(212)
        plt.title('Input')
        plt.plot(t, inj.vector * 1e9, label='injected (nA)')
        #plt.plot(t, gK.vector * 1e6, label='K')
        #plt.plot(t, gNa.vector * 1e6, label='Na')
        plt.legend()
        plt.figure()
        test_channel_gates()
        plt.show()
        plt.close()
        
    
if __name__ == '__main__':
    
    nogui = '-nogui' in sys.argv
    
    run(nogui)
    



# -*- coding: utf-8 -*-
# hdfutil.py ---
#
# Filename: hdfutil.py
# Description:
# Author:
# Maintainer:
# Created: Thu Aug 23 17:34:55 2012 (+0530)
# Version:
# Last-Updated: Mon Sep  3 17:55:03 2012 (+0530)
#           By: subha
#     Update #: 618
# URL:
# Keywords:
# Compatibility:
#
#

# Commentary:
#
# Utility function to save data in hdf5 format.
#
# In this utility we are trying to address a serialization
# problem. The ultimate goal is to be able to save a snapshot of
# complete simulator state in a portable format so that it can be
# loaded later to reach that state and continue from there.
#
# TODO: random number generators: the full RNG state has to be
# saved. MOOSE does not provide access to this at user level.
#
# TODO: what about internal variables? they affect the state of the
# simulation yet MOOSE does not allow access to these variables. Do we
# need a change in the API to make all internal variables accessible
# in a generic manner?
#
# TODO: How do we translate MOOSE tree to HDF5? MOOSE has vec and
# elements. vec is a container and each element belongs to an
# vec.
#
#                    em-0
#              el-00 el-01 el-02
#               /      |       \
#              /       |        \
#             em-1     em-2      em-3
#           el-10    el-20     el-30 el-31 el-32 el-33
#                     /                 \
#                    /                   \
#                   em-4                 em-5
#                el-40 el-41             el-50
#
#
# Serializing MOOSE tree structure into an HDF5 tree structure has
# some issues to be resolved.  Each vec is saved as a HDF
# group. All the elements inside it as a HDF dataset.  But the problem
# is that HDF datasets cannot have children. But in MOOSE the
# parent-child relation is opposite, each element can have one or more
# ematrices as children.
#
# Serializing MOOSE tree structure into HDF5 tables for each class.
# This is the approach I took initially. This is possibly more space
# saving.


# Change log:
#
#
#
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 3, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; see the file COPYING.  If not, write to
# the Free Software Foundation, Inc., 51 Franklin Street, Fifth
# Floor, Boston, MA 02110-1301, USA.
#
#

# Code:
from __future__ import print_function, division, absolute_import
try:
    from future_builtins import zip
except ImportError:
    pass
import moose
import numpy as np
import h5py as h5
import time
from collections import defaultdict

size_step=256


# maps cpp data type names to numpy data types
cpptonp = {
    'int': 'i4',
    'long': 'i8',
    'bool': 'b',
    'unsigned int': 'u4',
    'unsigned long': 'u8',
    'float': 'f4',
    'double': 'f8',
    'string': 'S1024',
    'Id': 'S1024',
    'ObjId': 'S1024',
    }

dtype_table = {}
# dims allows only 3 dims for the time being
em_dtype = np.dtype([('path', 'S1024'), ('class', 'S64'), ('dims', 'i4', (3,))])

def get_rec_dtype(em):
    bad_fields = []
    # Check if already have data type information for this class.
    # Create it otherwise.
    if em.className in dtype_table:
        dtype = dtype_table[em.className]
    else:
        print('Creating entries for class:', em.className)
        fielddict = moose.getFieldDict(em.className, 'valueFinfo')
        print(fielddict)

        # If we do not have the type of this field in cpp-np data
        # type map, skip it. We can handle vectors as nested
        # table, but the problem with that is numpy requires all
        # entries to have a fixed size for this table. Since
        # vectors can have arbitrary length, we cannot get such a
        # fixed size without checking the size of the vector field
        # in all elements.
        fields = [(fieldname, cpptonp[ftype]) # [('path', 'S1024')]
                  for fieldname, ftype in sorted(fielddict.items())
                  if ftype in cpptonp]
        dtype_table[em.className] = np.dtype(fields)
        return dtype_table[em.className]

def save_dataset(classname, rec, dtype, hdfnode):
    """Saves the data from rec into dataset"""
    # Check if there is a dataset for this class. Create it
    # otherwise.
    if len(rec) == 0:
        return
    if classname in hdfnode:
        ds = hdfnode[classname]
        oldlen = ds.len()
        newlen = oldlen + len(rec)
        ds.resize(newlen)
    else:
        ds = hdfnode.create_dataset(classname,
                                    shape=(len(rec),),
                                    dtype=dtype,
                                    compression='gzip',
                                    compression_opts=6)
    ds[oldlen:newlen] = np.rec.array(rec, dtype=dtype)

def savetree(moosenode, hdfnode):
    """Dump the MOOSE element tree rooted at moosenode as datasets
    under hdfnode."""
    # Keep track of total number of ematrices seen in each class.
    obj_count = defaultdict(int)
    # Buffer the last `size_step` object records for each class in
    # this array.
    obj_rec = defaultdict(list)
    em_rec = []
    hdfnode.attr['path'] = moosenode.path
    elements = hdfnode.create_group('elements')
    for em in moose.wildcardFind(moosenode.path+'/##'):
        em_rec.append((em.path, em.className, em.shape))
        dtype = get_rec_dtype(em)
        for obj in em:
            fields = []
            for fname in dtype.names:
                f = obj.getField(fname)
                if isinstance(f, moose.vec) or isinstance(f, moose.element):
                    fields.append(f.path)
                else:
                    fields.append(f)
            obj_rec[em.className].append(fields)
            obj_count[em.className] += 1
            if obj_count[em.className] % size_step == 0:
                save_dataset(em.className, obj_rec[em.className], dtype, elements)
                obj_rec[em.className][:] = [] # clear the records after saving
    # now save the remaining records (length < size_step)
    for classname, rec in list(obj_rec.items()):
        save_dataset(classname, rec, dtype_table[classname], hdfnode)
    vec = hdfnode.create_dataset('vec', shape=(len(em_rec),), dtype=em_dtype)
    vec[:] = em_rec

def loadtree(hdfnode, moosenode):
    """Load the element tree saved under the group `hdfnode` into `moosenode`"""
    pathclass = {}
    basepath = hdfnode.attr['path']
    if basepath != '/':
        basepath = basepath + '/'
    emdata = hdfnode['vec'][:]
    sorted_paths = sorted(emdata['path'], key=lambda x: x.count('/'))
    dims = dict(emdata['path', 'dims'])
    classes = dict(emdata['path', 'class'])
    current = moose.getCwe()
    moose.setCwe(moosenode)
    # First create all the ematrices
    for path in sorted_paths:
        rpath = path[len(basepath):]
        classname = classes[path]
        shape = dims[path]
        em = moose.vec(rpath, shape, classname)
    wfields = {}
    for cinfo in moose.element('/classes').children:
        cname = cinfo[0].name
        wfields[cname] = [f[len('set_'):] for f in moose.getFieldNames(cname, 'destFinfo')
                          if f.startswith('set_') and f not in ['set_this', 'set_name', 'set_lastDimension', 'set_runTime'] and not f.startswith('set_num_')]
        for key in hdfnode['/elements']:
            dset = hdfnode['/elements/'][key][:]
            fieldnames = dset.dtype.names
            for ii in range(len(dset)):
                obj = moose.element(dset['path'][ii][len(basepath):])
                for f in wfields[obj.className]:
                    obj.setField(f, dset[f][ii])



def savestate(filename=None):
    """Dump the state of MOOSE in an hdf5 file.

    The file will have a data set for each class.
    Each such dataset will be a column of field values.

    This function needs more work on moose serialization.
    """
    if filename is None:
        filename = 'moose_session_' + time.strftime('%Y%m%d_%H%M%S') + '.hdf5'
    with h5.File(filename, 'w') as fd:
        root = fd.create_group('/elements')
        meta = fd.create_group('/metadata')
        typeinfo_dataset = meta.create_dataset('typeinfo', shape=(size_step,), dtype=[('path', 'S1024'), ('class', 'S64'), ('dims', 'S64'), ('parent', 'S1024')], compression='gzip', compression_opts=6)
        typeinfo = []
        class_dataset_dict = {}
        class_count_dict = {}
        class_array_dict = {}
        objcount = 0
        for obj in moose.wildcardFind("/##"):
            if obj.path.startswith('/Msg') or obj.path.startswith('/class') or obj.className == 'Table' or obj.className == 'TableEntry':
                continue
            print('Processing:', obj.path, obj.className)
            typeinfo.append((obj.path, obj.className, str(obj.shape), obj[0].parent.path))
            objcount += 1
            if len(typeinfo) == size_step:
                typeinfo_dataset.resize(objcount)
                typeinfo_dataset[objcount - size_step: objcount] = np.rec.array(typeinfo, typeinfo_dataset.dtype)
                typeinfo = []
            # If we do not yet have dataset for this class, create one and keep it in dict
            if obj.className not in class_dataset_dict:
                print('Creating entries for class:', obj.className)
                fielddict = moose.getFieldDict(obj.className, 'valueFinfo')
                print(fielddict)
                keys = sorted(fielddict)
                fields = [] # [('path', 'S1024')]
                for fieldname in keys:
                    ftype = fielddict[fieldname]
                    if ftype in cpptonp:
                        fields.append((fieldname, cpptonp[ftype]))
                    elif ftype == 'Id' or ftype == 'ObjId':
                        fields.append((fieldname, 'S1024'))
                # print fields
                ds = root.create_dataset(obj.className, shape=(size_step,), dtype=fields, compression='gzip', compression_opts=6)
                class_dataset_dict[obj.className] = ds
                class_array_dict[obj.className] = []
                class_count_dict[obj.className] = 0
            # Lookup the dataset for the class this object belongs to
            ds = class_dataset_dict[obj.className]
            for entry in obj:
                fields = []
                print(entry.path, end=' ')
                for f in ds.dtype.names:
                    print('getting field:', f)
                    entry.getField(f)
                fields = [f.path if isinstance(f, moose.vec) or isinstance(f, moose.element) else f for f in fields]
                class_array_dict[obj.className].append(fields)
                # print 'fields:'
                # print fields
                # print 'length:', len(class_array_dict[obj.className])
                class_count_dict[obj.className] += 1
                if class_count_dict[obj.className] == size_step:
                    oldlen = ds.len()
                    if oldlen <= class_count_dict[obj.className]:
                        ds.resize((class_count_dict[obj.className]))
                    ds[oldlen: class_count_dict[obj.className]] = np.rec.array(class_array_dict[obj.className], dtype=ds.dtype)
                    class_array_dict[obj.className] = []
        for classname in class_array_dict:
            ds = class_dataset_dict[classname]
            ds.resize((class_count_dict[classname], ))
            if len(class_array_dict[classname]) > 0:
                start = class_count_dict[classname] - len(class_array_dict[classname])
                ds[start:] = np.rec.array(class_array_dict[classname], dtype=ds.dtype)

        if len(typeinfo) > 0:
            typeinfo_dataset.resize((objcount,))
            typeinfo_dataset[objcount-len(typeinfo): objcount] = np.rec.array(typeinfo, dtype=typeinfo_dataset.dtype)

def restorestate(filename):
    wfields = {}
    for cinfo in moose.element('/classes').children:
        cname = cinfo[0].name
        wfields[cname] = [f[len('set_'):] for f in moose.getFieldNames(cname, 'destFinfo')
                          if f.startswith('set_') and f not in ['set_this', 'set_name', 'set_lastDimension', 'set_runTime'] and not f.startswith('set_num_')]
    with h5.File(filename, 'r') as fd:
        typeinfo = fd['/metadata/typeinfo'][:]
        classdict = {}
        dimsdict = dict(zip(typeinfo['path'], typeinfo['dims']))
        classdict = dict(zip(typeinfo['path'], typeinfo['class']))
        parentdict = dict(zip(typeinfo['path'], typeinfo['parent']))
        sorted_paths = sorted(typeinfo['path'], key=lambda x: x.count('/'))
        for path in sorted_paths:
            name = path.rpartition('/')[-1].partition('[')[0]
            moose.vec(parentdict[path] + '/' + name, eval(dimsdict[path]), classdict[path])
        for key in fd['/elements']:
            dset = fd['/elements/'][key][:]
            fieldnames = dset.dtype.names
            for ii in range(len(dset)):
                obj = moose.element(dset['path'][ii])
                for f in wfields[obj.className]:
                    obj.setField(f, dset[f][ii])

#
# hdfutil.py ends here



# Modified from moose-example/synapse_tutorial.py ---

import moose
import random  
import numpy as np

def test_synapse():
    """
    In this example we walk through creation of a vector of IntFire
    elements and setting up synaptic connection between them. Synapse on
    IntFire elements is an example of ElementField - elements that do not
    exist on their own, but only as part of another element. This example
    also illustrates various operations on `vec` objects and
    ElementFields.
    """

    size = 1024  # number of IntFire objects in a vec
    delayMin = 0
    delayMax = 4
    thresh = 0.8
    weightMax = 0.5

    # The above sets the constants we shall use in this example. Now we create
    # a vector of IntFire elements of size `size`.
    net = moose.IntFire("/network", size)

    # This creates a `vec` of `IntFire`  elements of size 1024 and returns the
    # first `element`, i.e. "/network[0]".
    net = moose.element("/network[0]")

    # You need now to provide synaptic input to the network
    synh = moose.SimpleSynHandler("/network/synh", size)

    # These need to be connected to the nodes in the network
    moose.connect(synh, "activationOut", net, "activation", "OneToOne")

    # You can access the underlying vector of elements using the `vec` field on
    # any element. This is very useful for vectorized field access:
    _vm = [thresh / 2.0] * size
    net.vec.Vm = _vm
    assert np.allclose(net.vec.Vm, _vm)

    # The right part of the assigment creates a Python list of length `size`
    # with each element set to `thresh/2.0`, which is 0.4. You can index into
    # the `vec` to access individual elements' field:

    print(net.vec[1].Vm)
    assert net.vec[1].Vm == 0.4

    # `SimpleSynHandler` class has an `ElementField` called `synapse`. It is
    # just like a `vec` above in terms of field access, but by default its size
    # is 0.

    assert len(synh.synapse) == 0
    print(len(synh.synapse))

    # To actually create synapses, you can explicitly assign the `num` field of
    # this, or set the `numSynapses` field of the `IntFire` element. There are
    # some functions which can implicitly set the size of the `ElementField`.

    synh.numSynapses = 3
    assert len(synh.synapse) == 3
    print(len(synh.synapse))

    synh.numSynapses = 4
    print(synh.synapse, 111)
    assert len(synh.synapse) == 4, (4, len(synh.synapse))
    print(len(synh.synapse))

    # Now you can index into `net.synapse` as if it was an array.
    print("Before:", synh.synapse[0].delay)
    assert synh.synapse[0].delay == 0.0
    synh.synapse[0].delay = 1.0
    assert synh.synapse[0].delay == 1.0
    print("After:", synh.synapse[0].delay)

    # You could do the same vectorized assignment as with `vec` directly:

    syns = synh.synapse.vec
    syns.weight = [0.2] * len(syns)
    assert np.allclose(syns.weight, 0.2), syns.weight
    print(syns.weight)

    # You can create the synapses and assign the weights and delays using loops:
    for syn in synh.vec:
        syn.numSynapses = random.randint(1, 10) 

        # create synapse fields with random size between 1 and 10, end points
        # included. Below is one (inefficient) way of setting the individual weights of
        # the elements in 'synapse'
        syns  = syn.synapse
        for ii in range(len(syns)):
            syns[ii].weight = random.random() * weightMax

        # This is a more efficient way - rhs of `=` is list comprehension in
        # Python and rather fast.
        syns.delay = [
            delayMin + random.random() * delayMax for ii in range(len(syn.synapse))
        ]

        # An even faster way will be to use numpy.random.rand(size) which
        # produces array of random numbers uniformly distributed between 0 and
        # 1
        syns.delay = delayMin + np.random.rand(len(syn.synapse)) * delayMax

    # Now display the results, we use slice notation on `vec` to show the
    # values of delay and weight for the first 5 elements in `/network`
    for syn in synh.vec[:5]:
        print("Delays for synapses on ", syn.path, ":", syn.synapse.vec.delay)
        print("Weights for synapses on ", syn.path, ":", syn.synapse.vec.weight)

if __name__ == "__main__":
    test_synapse()



import os
import numpy as np
import moose
print( 'Using moose from %s' % moose.__file__ )
import time
os.environ['MOOSE_SHOW_PROFILING_INFO'] = '1'

def test_ksolver_parallel( nthreads = 4 ):
    """
    This example implements a reaction-diffusion like system which is
    bistable and propagates losslessly. It is based on the NEURON example 
    rxdrun.py, but incorporates more compartments and runs for a longer time.
    The system is implemented as a hybrid of a reaction and a function which
    sets its rates. Please see rxdFuncDiffusion.py for a variant that uses
    just a function object to set up the system.
    """

    dt = 0.1

    # define the geometry
    compt = moose.CylMesh( '/cylinder' )
    compt.r0 = compt.r1 = 1
    #  compt.diffLength = 1e-6
    compt.x1 = 1e-2
    assert compt.numDiffCompts == int(compt.x1/compt.diffLength), ( 
            compt.numDiffCompts, compt.x1 / compt.diffLength )
    print( 'No of compartment %d' % compt.numDiffCompts )

    #define the molecule. Its geometry is defined by its parent volume, cylinder
    c = moose.Pool( '/cylinder/pool' )
    c.diffConst = 1 # define diffusion constant
    # There is an implicit reaction substrate/product. MOOSE makes it explicit.
    buf = moose.BufPool( '/cylinder/buf' )
    buf.nInit = 1

    # The reaction is something entirely peculiar, not a chemical thing.
    reaction = moose.Reac( '/cylinder/reac' )
    reaction.Kb = 0

    # so here we set up a function calculation to do the same thing.
    func = moose.Function( '/cylinder/reac/func' )
    func.expr = "(1 - x0) * (0.3 - x0)"
    func.x.num = 1 #specify number of input variables.

    #Connect the reaction to the pools
    moose.connect( reaction, 'sub', c, 'reac' )
    moose.connect( reaction, 'prd', buf, 'reac' )

    #Connect the function to the reaction
    moose.connect( func, 'valueOut', reaction, 'setNumKf' )

    #Connect the molecules to the func
    moose.connect( c, 'nOut', func.x[0], 'input' )

    #Set up solvers
    ksolve = moose.Ksolve( '/cylinder/ksolve' )
    ksolve.numThreads = nthreads
    dsolve = moose.Dsolve( '/cylinder/dsolve' )
    stoich = moose.Stoich( '/cylinder/stoich' )
    stoich.compartment = compt
    stoich.ksolve = ksolve
    stoich.dsolve = dsolve
    stoich.reacSystemPath = '/cylinder/##'
    assert stoich.reacSystemPath == '/cylinder/##'
    for i in range( 10, 18 ):
        moose.setClock( i, dt )

    #initialize
    x = np.arange( 0, compt.x1, compt.diffLength )
    assert len(c.vec) == 10000, len(c.vec)
    nInit = [(float(q < 0.2) * compt.x1) for q in x]
    c.vec.nInit = nInit
    assert np.allclose(c.vec.nInit, nInit), (c.vec.nInit, nInit)

    expected = [ (0.01, 0.0)
            , (2.6704795776286974e-07, 1.2678976830753021e-17)
            , (8.167639617309419e-14, 3.8777269301457245e-24)
            , (2.498062905267963e-20, 1.1860363878961374e-30)
            , (7.64029581501609e-27, 3.6273808003690943e-37)
            ]

    # Run and plot it.
    moose.reinit()
    updateDt = 50
    runtime = updateDt * 4
    yvec = c.vec.n
    u1, m1 = np.mean( yvec ), np.std( yvec )
    print(u1, m1)
    assert np.isclose((u1, m1), expected[0], atol=1e-5).all(), ((u1, m1), expected[0])
    t1 = time.time()
    for i, t in enumerate(range( 0, runtime-1, updateDt)):
        moose.start(updateDt)
        yvec = c.vec.n
        u1, m1 = np.mean(yvec), np.std(yvec)
        print(u1, m1)
        np.isclose( (u1,m1), expected[i+1], atol=1e-5 ).all(), expected[i+1]
    return time.time() - t1

def main(nT):
    return test_ksolver_parallel(nT)

if __name__ == '__main__':
    import multiprocessing
    import sys
    nT = max(1, int(multiprocessing.cpu_count()))
    if len(sys.argv) > 1:
        nT = int(sys.argv[1])
    t2 = main( nT )
    print( 'With threads=%d, Time=%g' % (nT, t2))
